//******************************************************************************************
//* Poincar Hyperbolic Tiling, by Paul Nylander, bugman123.com, 6/19/05
//* runtime: 5 seconds
//******************************************************************************************

camera{location 2*y look_at 0}
light_source{2*y,1}

//Basic Functions
#macro square(X) X*X #end

//Complex Functions
#macro Complex(X,Y) <X,Y> #end
#declare I=Complex(0,1);
#macro Re(Z) Z.x #end
#macro Im(Z) Z.y #end
#macro Conjugate(Z) Complex(Re(Z),-Im(Z)) #end
#macro Abs(Z) vlength(Z) #end
#macro Arg(Z) atan2(Im(Z),Re(Z)) #end
#macro Sqr(Z) Complex(square(Re(Z))-square(Im(Z)),2*Re(Z)*Im(Z)) #end
#macro Pow(Z,n) #local r=Abs(Z); #if(r=0) #local z2=Complex(0,0); #else #local theta=n*Arg(Z); #local z2=pow(r,n)*Complex(cos(theta),sin(theta)); #end z2 #end
#macro Sqrt(Z) Pow(Z,1/2) #end
#macro Mult(z1,z2) Complex(Re(z1)*Re(z2)-Im(z1)*Im(z2),Im(z1)*Re(z2)+Re(z1)*Im(z2)) #end
#macro Div(z1,z2) Mult(z1,Pow(z2,-1)) #end
#macro Exp(Z) exp(Re(Z))*Complex(cos(Im(Z)),sin(Im(Z))) #end

//22 Matrix Functions
#macro Matrix(a,b,c,d) array[2][2]{{a,b},{c,d}} #end
#macro Conjugate2(A) Matrix(Conjugate(A[0][0]),Conjugate(A[0][1]),Conjugate(A[1][0]),Conjugate(A[1][1])) #end
#macro Det(A) Mult(A[0][0],A[1][1])-Mult(A[0][1],A[1][0]) #end
#macro Mult2(Z,A) Matrix(Mult(A[0][0],Z),Mult(A[0][1],Z),Mult(A[1][0],Z),Mult(A[1][1],Z)) #end
#macro Div2(A,Z) Matrix(Div(A[0][0],Z),Div(A[0][1],Z),Div(A[1][0],Z),Div(A[1][1],Z)) #end
#macro Inverse(A) Div2(Matrix(A[1][1],-A[0][1],-A[1][0],A[0][0]),Det(A)) #end
#macro Dot(A1,A2) Matrix(Mult(A1[0][0],A2[0][0])+Mult(A1[0][1],A2[1][0]),Mult(A1[0][0],A2[0][1])+Mult(A1[0][1],A2[1][1]),Mult(A2[0][0],A1[1][0])+Mult(A2[1][0],A1[1][1]),Mult(A2[0][1],A1[1][0])+Mult(A1[1][1],A2[1][1])) #end
#macro Reflect(C,M) Dot(Dot(M,Conjugate2(C)),M) #end
#macro ToMatrix(Z,r) Matrix(Z/r,Complex(r-(square(Re(Z))+square(Im(Z)))/r,0),Complex(1/r,0),-Conjugate(Z)/r) #end
#macro zcen(A) Div(A[0][0],A[1][0]) #end
#macro rad(A) abs(1/Re(A[1][0])) #end
#macro nth(i,Alist) Matrix(Alist[i][0][0],Alist[i][0][1],Alist[i][1][0],Alist[i][1][1]) #end
#macro setnth(i,Alist,A) #local Alist[i][0][0]=A[0][0]; #local Alist[i][0][1]=A[0][1]; #local Alist[i][1][0]=A[1][0]; #local Alist[i][1][1]=A[1][1]; Alist #end
#macro join(Alist1,Alist2) #local n1=dimension_size(Alist1,1); #local n2=dimension_size(Alist2,1);
 #local Alist3=array[n1+n2][2][2]; #local i=0;
 #while(i<n1) #local Alist3=setnth(i,Alist3,nth(i,Alist1)); #local i=i+1; #end #local i=0;
 #while(i<n2) #local Alist3=setnth(n1+i,Alist3,nth(i,Alist2)); #local i=i+1; #end Alist3
#end

//Calculations
#macro index(g,i) int(3*pow(2,g-1)-3+i) #end
#declare R=sqrt(3); #declare gmax=5; #declare Tiles=array[index(gmax+1,0)][2][2];
#declare Tiles=setnth(0,Tiles,ToMatrix(Complex(R,1),R));
#declare Tiles=setnth(1,Tiles,ToMatrix(Complex(-R,1),R));
#declare Tiles=setnth(2,Tiles,ToMatrix(-2*I,R));
#declare i=0;
#while(i<6)
 #declare Tiles=setnth(index(2,i),Tiles,Reflect(nth((mod(i,2)=0?(i<2?1:0):(i>3?1:2)),Tiles),nth(int(i/2),Tiles)));
 #declare i=i+1;
#end
#declare g=2;
#while(g<gmax) #declare i=0;
 #while(i<index(g,3))
  #declare Tiles=setnth(index(g+1,2*i),Tiles,Reflect(nth(index(g-1,floor(i/2)),Tiles),nth(index(g,i),Tiles)));
  #declare Tiles=setnth(index(g+1,2*i+1),Tiles,Reflect(nth(index(g,2*floor(i/2)+mod(i+1,2)),Tiles),nth(index(g,i),Tiles)));
  #declare i=i+1;
 #end
 #declare g=g+1;
#end

//Draw Chains
#declare rtube=0.01;
#macro DrawChain(A) #local Z=zcen(A);
 intersection{torus{rad(A),rtube translate <Re(Z),0,Im(Z)>} cylinder{-y,y,1} pigment{rgb 1}}
#end
torus{1,rtube pigment{rgb 1}}
disc{0,y,1 pigment{rgbt <1,0,1,0.5>}}
#declare i=0;
#while(i<dimension_size(Tiles,1)) DrawChain(nth(i,Tiles)) #declare i=i+1; #end
